import os
import subprocess
from glob import glob
from os.path import join
import jinja2

# Set torch HOME
os.environ["TORCH_HOME"] = os.path.expanduser("~/.cache/torch")

# Load a template named "README_template.jinja2"
templateLoader = jinja2.FileSystemLoader(searchpath="./")
templateEnv = jinja2.Environment(loader=templateLoader)
TEMPLATE_FILE = "README_template.md"

# Load examples
example_dict = {}
examples_dir = "quick_examples"
files = glob(join(examples_dir, "*.py"))

for i, file in enumerate(files):
    print(f"Running {file}")
    with open(file) as f:
        filename = os.path.basename(file).split(".")[0]
        example_dict[filename] = f.read()
        # execute py `file` and get output
        process = subprocess.run(["python", file], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        example_dict[filename + "_output"] = process.stdout.decode("utf-8")
        example_dict[filename + "_error"] = process.stderr.decode("utf-8")


readme = templateEnv.get_template(TEMPLATE_FILE).render(**example_dict)

# Write README.md
with open("README.md", "w") as f:
    f.write(readme)

print("README.md generated")



import torch
import torch.nn as nn
import numpy as np
from astra.torch.utils import train_fn
from astra.torch.models import AstraModel


class CustomModel(AstraModel):
    def __init__(self):
        super().__init__()
        self.linear = nn.Linear(2, 1)
        self.inp1_linear = nn.Linear(2, 1)

    def forward(self, x, inp1, fixed_bias):
        return self.linear(x) + self.inp1_linear(inp1) + fixed_bias


def custom_loss_fn(model_output, output, norm_factor):
    loss_fn = nn.MSELoss()
    loss_val = loss_fn(model_output, output)
    return loss_val / norm_factor


X = torch.randn(10, 2)
y = torch.randn(10, 1)
inp1 = torch.randn(10, 2)
bias = torch.randn(1)
norm_factor = torch.randn(1)

model = CustomModel()

optimizer = torch.optim.Adam(model.parameters(), lr=0.1)
(iter_losses, epoch_losses), state_dict_history = train_fn(
    model,
    input=X,  # Can be None if model.forward() does not require input
    model_kwargs={"inp1": inp1, "fixed_bias": bias},
    output=y,  # Can be None if loss_fn does not require output
    loss_fn=custom_loss_fn,
    loss_fn_kwargs={"norm_factor": norm_factor},
    optimizer=optimizer,
    epochs=5,
    shuffle=True,
    verbose=True,
    return_state_dict=True,
)

print("Epoch_losses", np.array(epoch_losses).round(2))



import torch
import torch.nn as nn
from torch.utils.data import TensorDataset, DataLoader

import numpy as np
from astra.torch.utils import train_fn
from astra.torch.models import CNNClassifier

torch.autograd.set_detect_anomaly(True)

X = torch.rand(100, 3, 28, 28)
y = torch.randint(0, 2, size=(200,)).reshape(100, 2).float()

model = CNNClassifier(
    image_dims=(28, 28), kernel_size=5, input_channels=3, conv_hidden_dims=[4], dense_hidden_dims=[2], n_classes=2
)

dataset = TensorDataset(X, y)
dataloader = DataLoader(dataset, batch_size=32, shuffle=True, num_workers=2)

# Let train_fn do the optimization for you
iter_losses, epoch_losses = train_fn(
    model,
    dataloader=dataloader,
    loss_fn=nn.CrossEntropyLoss(),
    lr=0.1,
    epochs=5,
)
print(np.array(epoch_losses).round(2))



import torch
from astra.torch.al.acquisitions.base import RandomAcquisition


class UniformRandomAcquisition(RandomAcquisition):
    def acquire_scores(self, pool_indices: torch.Tensor):
        # shape (pool_dim, )
        return torch.rand_like(pool_indices.float())



import pytest
import torch
from sklearn.metrics import (
    accuracy_score as sk_accuracy,
    recall_score as sk_recall,
    f1_score as sk_f1,
    precision_score as sk_precision,
)

from astra.torch.metrics import accuracy_score, precision_score, recall_score, f1_score

y = torch.randint(0, 2, (30,))
y_pred = torch.randint(0, 2, (30,))
target_accuracy = sk_accuracy(y.numpy(), y_pred.numpy())
target_precision_1 = sk_precision(y.numpy(), y_pred.numpy(), pos_label=1)
target_precision_0 = sk_precision(y.numpy(), y_pred.numpy(), pos_label=0)
target_recal_1 = sk_recall(y.numpy(), y_pred.numpy(), pos_label=1)
target_recal_0 = sk_recall(y.numpy(), y_pred.numpy(), pos_label=0)
target_f1_1 = sk_f1(y.numpy(), y_pred.numpy(), pos_label=1)
target_f1_0 = sk_f1(y.numpy(), y_pred.numpy(), pos_label=0)


def test_accuracy():
    assert accuracy_score(y_pred, y) == target_accuracy


def test_precision():
    assert precision_score(y_pred, y, pos_label=1) == target_precision_1
    assert precision_score(y_pred, y, pos_label=0) == target_precision_0


def test_recall():
    assert recall_score(y_pred, y, pos_label=1) == target_recal_1
    assert recall_score(y_pred, y, pos_label=0) == target_recal_0


def test_f1():
    assert f1_score(y_pred, y, pos_label=1) == target_f1_1
    assert f1_score(y_pred, y, pos_label=0) == target_f1_0



import torch
from torchvision.models import efficientnet_b0, EfficientNet_B0_Weights
from astra.torch.models import EfficientNetClassifier

# Pretrained model
model = EfficientNetClassifier(model=efficientnet_b0, weights=EfficientNet_B0_Weights.DEFAULT, n_classes=10)
# OR without pretrained weights
# model = EfficientNetClassifier(model=efficientnet_b0, weights=None, n_classes=10)

x = torch.rand(10, 3, 224, 224)
out = model(x)
print(out.shape)



import torch
from astra.torch.data import load_mnist, load_cifar_10


def test_load_mnist():
    dataset = load_mnist()
    assert dataset.data.shape == torch.Size([70000, 28, 28])
    assert dataset.targets.shape == torch.Size([70000])
    assert dataset.data.dtype == torch.float32
    assert dataset.targets.dtype == torch.int64
    assert dataset.classes == [
        "0 - zero",
        "1 - one",
        "2 - two",
        "3 - three",
        "4 - four",
        "5 - five",
        "6 - six",
        "7 - seven",
        "8 - eight",
        "9 - nine",
    ]


def test_load_cifar_10():
    dataset = load_cifar_10()
    assert dataset.data.shape == torch.Size([60000, 3, 32, 32])
    assert dataset.targets.shape == torch.Size([60000])
    assert dataset.data.dtype == torch.float32
    assert dataset.targets.dtype == torch.int64
    assert dataset.classes == [
        "airplane",
        "automobile",
        "bird",
        "cat",
        "deer",
        "dog",
        "frog",
        "horse",
        "ship",
        "truck",
    ]



from tqdm import tqdm
import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F

from torchvision.models.efficientnet import (
    EfficientNet as _EfficientNet,
)  # To prevent users from importing EfficientNet from torchvision
from torchvision.models.vision_transformer import VisionTransformer as _VisionTransformer  # To prevent ...
from torchvision.models.resnet import ResNet as _ResNet  # To prevent ...
from torchvision.models.densenet import DenseNet as _DenseNet  # To prevent ...
from torchvision.models.alexnet import AlexNet as _AlexNet  # To prevent ...
from torchvision.models._api import WeightsEnum


from torchvision.models import efficientnet_b0, EfficientNet_B0_Weights
from torchvision.models import vit_b_16, ViT_B_16_Weights
from torchvision.models import resnet18, ResNet18_Weights
from torchvision.models import densenet121, DenseNet121_Weights
from torchvision.models import alexnet, AlexNet_Weights

## Hotfix for https://github.com/pytorch/vision/issues/7744#issuecomment-1757321451
from torchvision.models._api import WeightsEnum
from torch.hub import load_state_dict_from_url

from typing import Tuple, Literal

from astra.torch.utils import get_model_device


def get_state_dict(self, *args, **kwargs):
    if "check_hash" in kwargs:
        kwargs.pop("check_hash")
    return load_state_dict_from_url(self.url, *args, **kwargs)


WeightsEnum.get_state_dict = get_state_dict
################################################################################


# Base class for all ASTRA models
class AstraModel(nn.Module):
    def predict(self, X=None, dataloader=None, batch_size=None, eval_mode: bool = True, verbose=True):
        device = get_model_device(self)
        if eval_mode:
            self.eval()
        else:
            self.train()

        assert (X is not None) ^ (dataloader is not None), "Either X or dataloader should be provided, but not both."

        if X is not None:
            if batch_size is None:
                batch_size = len(X)

            preds = []
            pbar = range(0, len(X), batch_size)
            if verbose:
                pbar = tqdm(pbar)
            for i in pbar:
                pred = self(X[i : i + batch_size].to(device))
                preds.append(pred)
            pred = torch.cat(preds)

            return pred

        elif dataloader is not None:
            preds = []
            pbar = dataloader
            if verbose:
                pbar = tqdm(pbar)
            for batch_input, _ in pbar:
                batch_input = batch_input.to(device)
                pred = self(batch_input)
                preds.append(pred)
            pred = torch.cat(preds)

            return pred

        else:
            raise ValueError("Either X or dataloader should be provided. This should never happen. Contact Zeel.")


# Classifier and regressor base classes
class Classifier(AstraModel):
    def __init__(self, featurizer, classifier):
        super().__init__()
        self.featurizer = featurizer
        self.classifier = classifier

    def forward(self, x):
        x = self.featurizer(x)
        x = self.classifier(x)
        return x

    def predict_class(self, X=None, dataloader=None, batch_size=None, eval_mode=True, verbose=True):
        logits = self.predict(X, dataloader, batch_size, eval_mode, verbose)
        return logits.argmax(dim=1)


class Regressor(AstraModel):
    def __init__(self, featurizer, regressor):
        super().__init__()
        self.featurizer = featurizer
        self.regressor = regressor

    def forward(self, x):
        x = self.featurizer(x)
        x = self.regressor(x)
        return x

    def rmse(self, X, y, batch_size=None, device=None, eval_mode: bool = True):
        y_pred = self.predict(X=X, batch_size=batch_size, device=device, eval_mode=eval_mode)

        if device is None:
            device = get_model_device(self)
        return F.mse_loss(y_pred, y.to(device), reduction="mean").sqrt().item()


# Multi-layer perceptron (MLP)
class MLP(AstraModel):
    """Multi-layer perceptron (MLP) with dropout and activation function."""

    def __init__(
        self,
        input_dim: int,
        hidden_dims: list,
        activation: nn.Module = nn.ReLU(),
        dropout: float = 0.0,
    ):
        super().__init__()

        self.input_dim = input_dim
        self.hidden_dims = hidden_dims

        # One time initialization layers
        self.activation = activation
        self.dropout = nn.Dropout(dropout)

        # Define input layer
        self.input_layer = nn.Linear(input_dim, self.hidden_dims[0])

        # Define hidden layers
        for i in range(1, len(self.hidden_dims)):
            hidden_layer = nn.Linear(self.hidden_dims[i - 1], self.hidden_dims[i])
            setattr(self, f"hidden_layer_{i}", hidden_layer)

    def forward(self, x):
        # Input layer
        x = self.input_layer(x)
        x = self.activation(x)
        x = self.dropout(x)

        # Hidden layers
        for i in range(1, len(self.hidden_dims)):
            x = getattr(self, f"hidden_layer_{i}")(x)
            x = self.activation(x)
            x = self.dropout(x)

        return x


class MLPClassifier(Classifier):
    def __init__(self, input_dim, hidden_dims, n_classes=2, activation=nn.ReLU(), dropout=0.0):
        """Multi-Layer Perceptron (MLP) classifier.

        Args:
            input_dim (int): Number of input features.
            hidden_dims (list): List of hidden dimensions.
            n_classes (int): Number of classes.
            activation (nn.Module, optional): Activation function. Defaults to nn.ReLU().
            dropout (float, optional): Dropout probability. Defaults to 0.0.
        """
        featurizer = MLP(input_dim, hidden_dims, activation, dropout)
        classifier = nn.Linear(hidden_dims[-1], n_classes)
        super().__init__(featurizer, classifier)


class MLPRegressor(Regressor):
    def __init__(self, input_dim, hidden_dims, output_dim: int = 1, activation=nn.ReLU(), dropout=0.0):
        """Multi-Layer Perceptron (MLP) regressor.

        Args:
            input_dim (int): Number of input features.
            hidden_dims (list): List of hidden dimensions.
            activation (nn.Module, optional): Activation function. Defaults to nn.ReLU().
            dropout (float, optional): Dropout probability. Defaults to 0.0.
        """
        featurizer = MLP(input_dim, hidden_dims, activation, dropout)
        regressor = nn.Linear(hidden_dims[-1], output_dim)
        super().__init__(featurizer, regressor)


# SIREN Regressor
class SIRENRegressor(MLPRegressor):
    """Sinusoidal Representation Network (SIREN). See https://arxiv.org/abs/2006.09661 (Implicit Neural Representations with Periodic Activation Functions by Sitzmann et al.)"""

    def __init__(
        self,
        input_dim: int,
        hidden_dims: list,
        output_dim: int = 1,
        activation_scale: float = 30.0,
        dropout: float = 0.0,
    ):
        super().__init__(
            input_dim,
            hidden_dims,
            output_dim,
            activation=lambda x: torch.sin(activation_scale * x),
            dropout=dropout,
        )

        # Initialize weights with SIREN initialization
        self._initialize(activation_scale)

    def _initialize(self, activation_scale):
        def first_layer_init(m):
            if hasattr(m, "weight"):
                input_size = m.weight.size(-1)
                m.weight.uniform_(-1 / input_size, 1 / input_size)

        def other_layer_init(m):
            if hasattr(m, "weight"):
                input_size = m.weight.size(-1)
                m.weight.uniform_(
                    -np.sqrt(6 / input_size) / activation_scale, np.sqrt(6 / input_size) / activation_scale
                )

        with torch.no_grad():
            self.apply(other_layer_init)
            self.featurizer.input_layer.apply(first_layer_init)


# CNN Models
class CNN(AstraModel):
    def __init__(
        self,
        kernel_size: int,
        input_channels: int,
        hidden_dims: list,
        activation: nn.Module = nn.ReLU(),
        adaptive_pooling: bool = False,
    ):
        super().__init__()

        self.hidden_dims = hidden_dims

        padding = kernel_size // 2

        # One time initialization layers
        self.activation = activation

        self.max_pool = nn.MaxPool2d(kernel_size=2, stride=2)

        self.input_layer = nn.Conv2d(input_channels, self.hidden_dims[0], kernel_size=kernel_size, padding=padding)

        # Define hidden layers
        for i in range(1, len(self.hidden_dims)):
            hidden_layer = nn.Conv2d(
                self.hidden_dims[i - 1],
                self.hidden_dims[i],
                kernel_size=kernel_size,
                padding=padding,
            )
            setattr(self, f"hidden_layer_{i}", hidden_layer)

        # Define aggregator
        if adaptive_pooling:
            aggregator = nn.AdaptiveAvgPool2d((1, 1))
        else:
            aggregator = nn.Identity()
        self.aggregator = aggregator

        self.flatten = nn.Flatten()

    def forward(self, x):
        x = self.input_layer(x)
        x = self.activation(x)
        x = self.max_pool(x)

        for i in range(1, len(self.hidden_dims)):
            x = getattr(self, f"hidden_layer_{i}")(x)
            x = self.activation(x)
            x = self.max_pool(x)

        # Aggregator
        x = self.aggregator(x)

        # Flatten
        x = self.flatten(x)

        return x


class CNNClassifier(Classifier):
    def __init__(
        self,
        image_dims: Tuple[int, int],
        kernel_size: int,
        input_channels: int,
        conv_hidden_dims: list,
        dense_hidden_dims: list,
        n_classes: int = 2,
        activation: nn.Module = nn.ReLU(),
        dropout: float = 0.0,
        adaptive_pooling: bool = False,
    ):
        """Convolutional Neural Network (CNN) classifier.

        Args:
            image_dim (int): Image dimension.
            kernel_size (int): Kernel size.
            n_channels (int): Number of input channels.
            hidden_dims (list): List of convolutional hidden dimensions.
            dense_hidden_dims (list): List of dense hidden dimensions.
            n_classes (int): Number of classes.
            activation (nn.Module, optional): Activation function. Defaults to nn.ReLU().
            dropout (float, optional): Dropout probability. Defaults to 0.0.
        """
        featurizer = CNN(kernel_size, input_channels, conv_hidden_dims, activation, adaptive_pooling)
        if adaptive_pooling:
            mlp_input_dim = conv_hidden_dims[-1]
        else:
            out_image_dim0 = image_dims[0] // 2 ** len(conv_hidden_dims)
            out_image_dim1 = image_dims[1] // 2 ** len(conv_hidden_dims)
            mlp_input_dim = conv_hidden_dims[-1] * out_image_dim0 * out_image_dim1
        classifier = MLPClassifier(mlp_input_dim, dense_hidden_dims, n_classes, activation, dropout)
        super().__init__(featurizer, classifier)


class CNNRegressor(Regressor):
    def __init__(
        self,
        image_dims: Tuple[int, int],
        kernel_size: int,
        input_channels: int,
        conv_hidden_dims: list,
        dense_hidden_dims: list,
        output_dim: int = 1,
        activation: nn.Module = nn.ReLU(),
        dropout: float = 0.0,
        adaptive_pooling: bool = False,
    ):
        """Convolutional Neural Network (CNN) regressor.

        Args:
            image_dim (int): Image dimension.
            kernel_size (int): Kernel size.
            n_channels (int): Number of input channels.
            hidden_dims (list): List of convolutional hidden dimensions.
            dense_hidden_dims (list): List of dense hidden dimensions.
            output_dim (int): Number of output features.
            activation (nn.Module, optional): Activation function. Defaults to nn.ReLU().
            dropout (float, optional): Dropout probability. Defaults to 0.0.
        """
        featurizer = CNN(kernel_size, input_channels, conv_hidden_dims, activation, adaptive_pooling)
        if adaptive_pooling:
            mlp_input_dim = conv_hidden_dims[-1]
        else:
            out_image_dim0 = image_dims[0] // 2 ** len(conv_hidden_dims)
            out_image_dim1 = image_dims[1] // 2 ** len(conv_hidden_dims)
            mlp_input_dim = conv_hidden_dims[-1] * out_image_dim0 * out_image_dim1
        regressor = MLPRegressor(mlp_input_dim, dense_hidden_dims, output_dim, activation, dropout)
        super().__init__(featurizer, regressor)


# ResNet
class ResNet(AstraModel):
    """ResNet. See https://arxiv.org/abs/1512.03385 (Deep Residual Learning for Image Recognition by He et al.)"""

    def __init__(
        self,
        model: _ResNet = resnet18,
        weights: WeightsEnum = ResNet18_Weights.DEFAULT,
        transform: bool = False,
    ):
        super().__init__()

        self.transform = weights.transforms() if transform else None
        self.resnet = model(weights=weights)
        self.resnet.fc = nn.Identity()
        self.flatten = nn.Flatten()

    def forward(self, x):
        if self.transform is not None:
            x = self.transform(x)
        x = self.resnet(x)
        x = self.flatten(x)
        return x


class ResNetClassifier(Classifier):
    def __init__(
        self,
        model: _ResNet = resnet18,
        weights: WeightsEnum = ResNet18_Weights.DEFAULT,
        n_classes: int = 2,
        transform: bool = False,
        dense_hidden_dims: list = [512],
        activation: nn.Module = nn.ReLU(),
        dropout: float = 0.0,
    ):
        """ResNet classifier.

        Args:
            model (_ResNet, optional): _description_. Defaults to resnet18.
            weights (WeightsEnum, optional): _description_. Defaults to ResNet18_Weights.DEFAULT.
            transform (bool, optional): _description_. Defaults to False.
            dense_hidden_dims (list, optional): _description_. Defaults to [512].
            activation (nn.Module, optional): _description_. Defaults to nn.ReLU().
            n_classes (int, optional): _description_. Defaults to 1.
        """
        featurizer = ResNet(model, weights, transform)
        classifier = MLPClassifier(512, dense_hidden_dims, n_classes, activation, dropout)
        super().__init__(featurizer, classifier)


class ResNetRegressor(Regressor):
    def __init__(
        self,
        model: _ResNet = resnet18,
        weights: WeightsEnum = ResNet18_Weights.DEFAULT,
        output_dim: int = 1,
        transform: bool = False,
        dense_hidden_dims: list = [512],
        activation: nn.Module = nn.ReLU(),
        dropout: float = 0.0,
    ):
        """ResNet regressor.

        Args:
            model (_ResNet, optional): _description_. Defaults to resnet18.
            weights (WeightsEnum, optional): _description_. Defaults to ResNet18_Weights.DEFAULT.
            transform (bool, optional): _description_. Defaults to False.
            dense_hidden_dims (list, optional): _description_. Defaults to [512].
            activation (nn.Module, optional): _description_. Defaults to nn.ReLU().
            output_dim (int, optional): _description_. Defaults to 1.
        """
        featurizer = ResNet(model, weights, transform)
        regressor = MLPRegressor(512, dense_hidden_dims, output_dim, activation, dropout)
        super().__init__(featurizer, regressor)


# EfficientNet
class EfficientNet(AstraModel):
    """EfficientNet. See https://arxiv.org/abs/1905.11946 (EfficientNet: Rethinking Model Scaling for Convolutional Neural Networks by Tan et al.)"""

    def __init__(
        self,
        model: _EfficientNet = efficientnet_b0,
        weights: WeightsEnum = EfficientNet_B0_Weights.DEFAULT,
        transform: bool = False,
    ):
        super().__init__()

        self.transform = weights.transforms() if transform else None
        self.efficientnet = model(weights=weights)
        self.efficientnet.classifier = nn.Identity()
        self.flatten = nn.Flatten()

    def forward(self, x):
        if self.transform is not None:
            x = self.transform(x)
        x = self.efficientnet(x)
        x = self.flatten(x)
        return x


class EfficientNetClassifier(Classifier):
    def __init__(
        self,
        model: _EfficientNet = efficientnet_b0,
        weights: WeightsEnum = EfficientNet_B0_Weights.DEFAULT,
        n_classes: int = 2,
        transform: bool = False,
        dense_hidden_dims: list = [512],
        activation: nn.Module = nn.ReLU(),
        dropout: float = 0.0,
    ):
        """EfficientNet classifier.

        Args:
            model (_EfficientNet, optional): _description_. Defaults to efficientnet_b0.
            weights (WeightsEnum, optional): _description_. Defaults to EfficientNet_B0_Weights.DEFAULT.
            transform (bool, optional): _description_. Defaults to False.
            dense_hidden_dims (list, optional): _description_. Defaults to [512].
            activation (nn.Module, optional): _description_. Defaults to nn.ReLU().
            n_classes (int, optional): _description_. Defaults to 1.
        """
        featurizer = EfficientNet(model, weights, transform)
        # Efficient uses AdaptiveAvgPool2d so output_dim is always 1280
        classifier = MLPClassifier(1280, dense_hidden_dims, n_classes, activation, dropout)
        super().__init__(featurizer, classifier)


class EfficientNetRegressor(Regressor):
    def __init__(
        self,
        model: _EfficientNet = efficientnet_b0,
        weights: WeightsEnum = EfficientNet_B0_Weights.DEFAULT,
        output_dim: int = 1,
        transform: bool = False,
        dense_hidden_dims: list = [512],
        activation: nn.Module = nn.ReLU(),
        dropout: float = 0.0,
    ):
        """EfficientNet regressor.

        Args:
            model (_EfficientNet, optional): _description_. Defaults to efficientnet_b0.
            weights (WeightsEnum, optional): _description_. Defaults to EfficientNet_B0_Weights.DEFAULT.
            transform (bool, optional): _description_. Defaults to False.
            dense_hidden_dims (list, optional): _description_. Defaults to [512].
            activation (nn.Module, optional): _description_. Defaults to nn.ReLU().
            output_dim (int, optional): _description_. Defaults to 1.
        """
        featurizer = EfficientNet(model, weights, transform)
        # Efficient uses AdaptiveAvgPool2d so output_dim is always 1280
        regressor = MLPRegressor(1280, dense_hidden_dims, output_dim, activation, dropout)
        super().__init__(featurizer, regressor)


class DenseNet(AstraModel):
    """DenseNet. See https://arxiv.org/abs/1608.06993 (Densely Connected Convolutional Networks by Huang et al.)"""

    def __init__(
        self,
        model: _DenseNet = densenet121,
        weights: WeightsEnum = DenseNet121_Weights.DEFAULT,
        transform: bool = False,
    ):
        super().__init__()

        self.transform = weights.transforms() if transform else None
        self.densenet = model(weights=weights)
        self.densenet.classifier = nn.Identity()

    def forward(self, x):
        if self.transform is not None:
            x = self.transform(x)
        x = self.densenet(x)
        return x


class DenseNetClassifier(Classifier):
    def __init__(
        self,
        model: _DenseNet = densenet121,
        weights: WeightsEnum = DenseNet121_Weights.DEFAULT,
        n_classes: int = 2,
        transform: bool = False,
        dense_hidden_dims: list = [512],
        activation: nn.Module = nn.ReLU(),
        dropout: float = 0.0,
    ):
        """DenseNet classifier.

        Args:
            model (_DenseNet, optional): _description_. Defaults to densenet121.
            weights (WeightsEnum, optional): _description_. Defaults to DenseNet121_Weights.DEFAULT.
            transform (bool, optional): _description_. Defaults to False.
            dense_hidden_dims (list, optional): _description_. Defaults to [512].
            activation (nn.Module, optional): _description_. Defaults to nn.ReLU().
            n_classes (int, optional): _description_. Defaults to 1.
        """
        featurizer = DenseNet(model, weights, transform)
        # DenseNet uses AdaptiveAvgPool2d so output_dim is always 1024
        classifier = MLPClassifier(1024, dense_hidden_dims, n_classes, activation, dropout)
        super().__init__(featurizer, classifier)


class DenseNetRegressor(Regressor):
    def __init__(
        self,
        model: _DenseNet = densenet121,
        weights: WeightsEnum = DenseNet121_Weights.DEFAULT,
        output_dim: int = 1,
        transform: bool = False,
        dense_hidden_dims: list = [512],
        activation: nn.Module = nn.ReLU(),
        dropout: float = 0.0,
    ):
        """DenseNet regressor.

        Args:
            model (_DenseNet, optional): _description_. Defaults to densenet121.
            weights (WeightsEnum, optional): _description_. Defaults to DenseNet121_Weights.DEFAULT.
            transform (bool, optional): _description_. Defaults to False.
            dense_hidden_dims (list, optional): _description_. Defaults to [512].
            activation (nn.Module, optional): _description_. Defaults to nn.ReLU().
            output_dim (int, optional): _description_. Defaults to 1.
        """
        featurizer = DenseNet(model, weights, transform)
        # DenseNet uses AdaptiveAvgPool2d so output_dim is always 1024
        regressor = MLPRegressor(1024, dense_hidden_dims, output_dim, activation, dropout)
        super().__init__(featurizer, regressor)


class AlexNet(AstraModel):
    """AlexNet. See https://arxiv.org/abs/1404.5997 (ImageNet Classification with Deep Convolutional Neural Networks by Krizhevsky et al.)"""

    def __init__(
        self,
        model: _AlexNet = alexnet,
        weights: WeightsEnum = AlexNet_Weights.DEFAULT,
        transform: bool = False,
    ):
        super().__init__()

        self.transform = weights.transforms() if transform else None
        self.alexnet = model(weights=weights)
        self.alexnet.classifier._modules["6"] = nn.Identity()

    def forward(self, x):
        if self.transform is not None:
            x = self.transform(x)
        x = self.alexnet(x)
        return x


class AlexNetClassifier(Classifier):
    def __init__(
        self,
        model: _AlexNet = alexnet,
        weights: WeightsEnum = AlexNet_Weights.DEFAULT,
        n_classes: int = 2,
        transform: bool = False,
        dense_hidden_dims: list = [512],
        activation: nn.Module = nn.ReLU(),
        dropout: float = 0.0,
    ):
        """AlexNet classifier.

        Args:
            model (_AlexNet, optional): _description_. Defaults to alexnet.
            weights (WeightsEnum, optional): _description_. Defaults to AlexNet_Weights.DEFAULT.
            transform (bool, optional): _description_. Defaults to False.
            dense_hidden_dims (list, optional): _description_. Defaults to [512].
            activation (nn.Module, optional): _description_. Defaults to nn.ReLU().
            n_classes (int, optional): _description_. Defaults to 1.
        """
        featurizer = AlexNet(model, weights, transform)
        # AlexNet uses AdaptiveAvgPool2d so output_dim is always 4096
        classifier = MLPClassifier(4096, dense_hidden_dims, n_classes, activation, dropout)
        super().__init__(featurizer, classifier)


class AlexNetRegressor(Regressor):
    def __init__(
        self,
        model: _AlexNet = alexnet,
        weights: WeightsEnum = AlexNet_Weights.DEFAULT,
        output_dim: int = 1,
        transform: bool = False,
        dense_hidden_dims: list = [512],
        activation: nn.Module = nn.ReLU(),
        dropout: float = 0.0,
    ):
        """AlexNet regressor.

        Args:
            model (_AlexNet, optional): _description_. Defaults to alexnet.
            weights (WeightsEnum, optional): _description_. Defaults to AlexNet_Weights.DEFAULT.
            transform (bool, optional): _description_. Defaults to False.
            dense_hidden_dims (list, optional): _description_. Defaults to [512].
            activation (nn.Module, optional): _description_. Defaults to nn.ReLU().
            output_dim (int, optional): _description_. Defaults to 1.
        """
        featurizer = AlexNet(model, weights, transform)
        # AlexNet uses AdaptiveAvgPool2d so output_dim is always 4096
        regressor = MLPRegressor(4096, dense_hidden_dims, output_dim, activation, dropout)
        super().__init__(featurizer, regressor)


class ViT(AstraModel):
    def __init__(
        self,
        model: _VisionTransformer = vit_b_16,
        weights: WeightsEnum = ViT_B_16_Weights.DEFAULT,
        transform: bool = False,
    ):
        """Vision transformer (ViT). See https://arxiv.org/abs/2010.11929 (An Image is Worth 16x16 Words: Transformers for Image Recognition at Scale by Dosovitskiy et al.)"""
        super().__init__()

        self.transform = weights.transforms() if transform else None
        self.vit = model(weights=weights)
        self.vit.heads = nn.Identity()
        self.flatten = nn.Flatten()

    def forward(self, x):
        if self.transform is not None:
            x = self.transform(x)
        x = self.vit(x)
        x = self.flatten(x)
        return x


class ViTClassifier(Classifier):
    def __init__(
        self,
        model: _VisionTransformer = vit_b_16,
        weights: WeightsEnum = ViT_B_16_Weights.DEFAULT,
        n_classes: int = 2,
        transform: bool = False,
        dense_hidden_dims: list = [512],
        activation: nn.Module = nn.ReLU(),
        dropout: float = 0.0,
    ):
        """Vision transformer (ViT) classifier.

        Args:
            model (_VisionTransformer, optional): _description_. Defaults to vit_b_16.
            weights (WeightsEnum, optional): _description_. Defaults to ViT_B_16_Weights.DEFAULT.
            transform (bool, optional): _description_. Defaults to False.
            dense_hidden_dims (list, optional): _description_. Defaults to [512].
            activation (nn.Module, optional): _description_. Defaults to nn.ReLU().
            n_classes (int, optional): _description_. Defaults to 1.
        """
        featurizer = ViT(model, weights, transform)
        # ViT uses fixed input size of 224 so output_dim is always 768
        classifier = MLPClassifier(768, dense_hidden_dims, n_classes, activation, dropout)
        super().__init__(featurizer, classifier)


class ViTRegressor(Regressor):
    def __init__(
        self,
        model: _VisionTransformer = vit_b_16,
        weights: WeightsEnum = ViT_B_16_Weights.DEFAULT,
        output_dim: int = 1,
        transform: bool = False,
        dense_hidden_dims: list = [512],
        activation: nn.Module = nn.ReLU(),
        dropout: float = 0.0,
    ):
        """Vision transformer (ViT) regressor.

        Args:
            model (_VisionTransformer, optional): _description_. Defaults to vit_b_16.
            weights (WeightsEnum, optional): _description_. Defaults to ViT_B_16_Weights.DEFAULT.
            transform (bool, optional): _description_. Defaults to False.
            dense_hidden_dims (list, optional): _description_. Defaults to [512].
            activation (nn.Module, optional): _description_. Defaults to nn.ReLU().
            output_dim (int, optional): _description_. Defaults to 1.
        """
        featurizer = ViT(model, weights, transform)
        # ViT uses fixed input size of 224 so output_dim is always 768
        regressor = MLPRegressor(768, dense_hidden_dims, output_dim, activation, dropout)
        super().__init__(featurizer, regressor)



import numpy as np

import torch
import torch.nn as nn
from torch.utils.data import DataLoader

from astra.torch.al.strategies.base import Strategy
from astra.torch.al.acquisitions.base import DeterministicAcquisition
from astra.torch.al.errors import AcquisitionMismatchError

from typing import Sequence, Dict, Union


class DeterministicStrategy(Strategy):
    def __init__(
        self,
        acquisitions: Union[DeterministicAcquisition, Sequence[DeterministicAcquisition]],
        inputs: torch.Tensor,
        outputs: torch.Tensor,
    ):
        """Base class for query strategies

        Args:
            acquisitions: A sequence of acquisition functions.
            inputs: A tensor of inputs.
            outputs: A tensor of outputs.
        """
        super().__init__(acquisitions, inputs, outputs)

        for name, acquisition in self.acquisitions.items():
            if not isinstance(acquisition, DeterministicAcquisition):
                raise AcquisitionMismatchError(DeterministicAcquisition, acquisition)

    def query(
        self,
        net: nn.Module,
        pool_indices: torch.Tensor,
        context_indices: torch.Tensor = None,
        n_query_samples: int = 1,
        n_mc_samples: int = 10,
        batch_size: int = None,
    ) -> Dict[str, torch.Tensor]:
        """Monte Carlo query strategy

        Args:
            net: A neural network with dropout layers.
            pool_indices: The indices of the pool set.
            context_indices: This argument is ignored.
            n_query_samples: Number of samples to query.
            n_mc_samples: This argument is ignored.
            batch_size: Batch size for the data loader.

        Returns:
            best_indices: A dictionary of acquisition names and the corresponding best indices.
        """
        assert isinstance(pool_indices, torch.Tensor), f"pool_indices must be a torch.Tensor, got {type(pool_indices)}"

        if batch_size is None:
            batch_size = len(pool_indices)

        data_loader = DataLoader(self.dataset[pool_indices])

        # put the model on eval mode
        net.eval()

        with torch.no_grad():
            logits_list = []
            for x, _ in data_loader:
                logits = net(x)
                logits_list.append(logits)
            logits = torch.cat(logits_list, dim=1)  # (pool_dim, n_classes)

            best_indices = {}
            for acq_name, acquisition in self.acquisitions.items():
                scores = acquisition.acquire_scores(logits)
                index = torch.topk(scores, n_query_samples).indices
                selected_indices = pool_indices[index]
                best_indices[acq_name] = selected_indices

        return best_indices



import torch
import torch.nn as nn
import numpy as np
from astra.torch.utils import train_fn
from astra.torch.models import CNNClassifier

torch.autograd.set_detect_anomaly(True)

X = torch.rand(100, 3, 28, 28)
y = torch.randint(0, 2, size=(200,)).reshape(100, 2).float()

model = CNNClassifier(
    image_dims=(28, 28), kernel_size=5, input_channels=3, conv_hidden_dims=[4], dense_hidden_dims=[2], n_classes=2
)

# Let train_fn do the optimization for you
iter_losses, epoch_losses = train_fn(
    model, input=X, output=y, loss_fn=nn.CrossEntropyLoss(), lr=0.1, epochs=5, verbose=False
)
print(np.array(epoch_losses).round(2))

# OR

# Define your own optimizer

optimizer = torch.optim.Adam(model.parameters(), lr=0.1)
iter_losses, epoch_losses = train_fn(
    model,
    input=X,
    output=y,
    loss_fn=nn.MSELoss(),
    optimizer=optimizer,
    verbose=False,
    epochs=5,
)
print(np.array(epoch_losses).round(2))

# Get the state_dict of the model at each epoch

(iter_losses, epoch_losses), state_dict_history = train_fn(
    model,
    input=X,
    output=y,
    loss_fn=nn.MSELoss(),
    lr=0.1,
    epochs=5,
    verbose=False,
    return_state_dict=True,
)
print(np.array(epoch_losses).round(2))



import torch
from torchvision.models import vit_b_16, ViT_B_16_Weights
from astra.torch.models import ViTClassifier

model = ViTClassifier(vit_b_16, ViT_B_16_Weights.DEFAULT, n_classes=10)
x = torch.rand(10, 3, 224, 224)  # (batch_size, channels, h, w)
out = model(x)
print(out.shape)



import torch


def accuracy_score(y_pred, y):
    return (y_pred == y).float().mean()


def true_positives_score(y_pred, y, pos_label=1):
    return torch.logical_and(y_pred == pos_label, y == pos_label).sum()


def false_positives_score(y_pred, y, pos_label=1):
    return torch.logical_and(y_pred == pos_label, y != pos_label).sum()


def false_negatives_score(y_pred, y, pos_label=1):
    return torch.logical_and(y_pred != pos_label, y == pos_label).sum()


def precision_score(y_pred, y, pos_label=1):
    tp = true_positives_score(y_pred, y, pos_label)
    fp = false_positives_score(y_pred, y, pos_label)

    return tp / (tp + fp)


def recall_score(y_pred, y, pos_label=1):
    tp = true_positives_score(y_pred, y, pos_label)
    fn = false_negatives_score(y_pred, y, pos_label)

    return tp / (tp + fn)


def f1_score(y_pred, y, pos_label=1):
    tp = true_positives_score(y_pred, y, pos_label)
    fp = false_positives_score(y_pred, y, pos_label)
    fn = false_negatives_score(y_pred, y, pos_label)

    return tp / (tp + (fp + fn) / 2)



from astra.torch.models import CNNClassifier

cnn = CNNClassifier(
    image_dims=(32, 32),
    kernel_size=5,
    input_channels=3,
    conv_hidden_dims=[32, 64],
    dense_hidden_dims=[128, 64],
    n_classes=10,
)
print(cnn)



import numpy as np
import torch
import torch.nn as nn
from astra.torch.al.strategies.base import Strategy
from astra.torch.al.acquisitions.base import RandomAcquisition
from astra.torch.al.errors import AcquisitionMismatchError

from typing import Dict, Sequence, Union


class RandomStrategy(Strategy):
    def __init__(
        self,
        acquisitions: Union[RandomAcquisition, Sequence[RandomAcquisition]],
        inputs: torch.Tensor,
        outputs: torch.Tensor,
    ):
        """Random query strategy to randomly select samples.

        Args:
            acquisitions: A sequence of acquisition functions.
            inputs: A tensor of inputs.
            outputs: A tensor of outputs.
        """
        super().__init__(acquisitions, inputs, outputs)

        for name, acquisition in self.acquisitions.items():
            if not isinstance(acquisition, RandomAcquisition):
                raise AcquisitionMismatchError(RandomAcquisition, acquisition)

    def query(
        self,
        net: nn.Module,
        pool_indices: torch.Tensor,
        context_indices: torch.Tensor = None,
        n_query_samples: int = 1,
        n_mc_samples: int = None,
        batch_size: int = None,
    ) -> Dict[str, torch.Tensor]:
        """
        Args:
            net: This argument is ignored.
            pool_indices: The indices of the pool set.
            context_indices: This argument is ignored.
            n_query_samples: Number of samples to query.
            n_mc_samples: This argument is ignored.
            batch_size: This argument is ignored.

        Returns:
            best_indices: A dictionary of acquisition names and the corresponding best indices.
        """
        assert isinstance(pool_indices, torch.Tensor), f"pool_indices must be a torch.Tensor, got {type(pool_indices)}"

        best_indices = {}
        for acq_name, acquisition in self.acquisitions.items():
            scores = acquisition.acquire_scores(pool_indices)
            indices = torch.topk(scores, n_query_samples).indices
            selected_indices = pool_indices[indices]
            best_indices[acq_name] = selected_indices

        return best_indices



# Strategies
from astra.torch.al.strategies.base import Strategy
from astra.torch.al.strategies.deterministic import DeterministicStrategy
from astra.torch.al.strategies.diversity import DiversityStrategy
from astra.torch.al.strategies.ensemble import EnsembleStrategy
from astra.torch.al.strategies.mc import MCStrategy
from astra.torch.al.strategies.random import RandomStrategy

# Acquisition base classes
from astra.torch.al.acquisitions.base import (
    Acquisition,
    RandomAcquisition,
    MCAcquisition,
    EnsembleAcquisition,
    DiversityAcquisition,
    DeterministicAcquisition,
)

# Acquisition functions
from astra.torch.al.acquisitions.uniform_random import UniformRandomAcquisition



import pytest

import torch
from astra.torch.al import (
    UniformRandomAcquisition,
    RandomStrategy,
    EnsembleAcquisition,
    EnsembleStrategy,
    MCStrategy,
    MCAcquisition,
    DiversityStrategy,
    DiversityAcquisition,
    DeterministicAcquisition,
    DeterministicStrategy,
)
from astra.torch.al.errors import AcquisitionMismatchError


def test_acquisition_match():
    inputs = torch.rand(10, 3, 32, 32)
    outputs = torch.randint(0, 10, (10,))
    acquision = UniformRandomAcquisition()
    strategy = RandomStrategy(acquision, inputs, outputs)
    # If code reaches here, then the test passes


@pytest.mark.parametrize(
    "another_class", [EnsembleAcquisition, MCAcquisition, DiversityAcquisition, DeterministicAcquisition]
)
def test_acquisition_mismatch(another_class):
    inputs = torch.rand(10, 3, 32, 32)
    outputs = torch.randint(0, 10, (10,))

    class DummyAcquisition(another_class):
        def acquire_scores(self, logits):
            return logits

    another_acquisition = DummyAcquisition()

    # capture acquisition mismatch error
    with pytest.raises(AcquisitionMismatchError):
        strategy = RandomStrategy(another_acquisition, inputs, outputs)



import pytest
import torch
import torch.nn as nn
import torch.distributions as dist

from astra.torch.models import CNNClassifier, MLPRegressor
from astra.torch.utils import train_fn, count_params

device = torch.device("cuda" if torch.cuda.is_available() else "cpu")


@pytest.mark.parametrize(
    "model, expected_size",
    [
        (MLPRegressor(input_dim=2, hidden_dims=[3, 4], output_dim=3), 40),
        (
            CNNClassifier(
                image_dims=(23, 23),
                kernel_size=3,
                input_channels=3,
                conv_hidden_dims=[2, 3],
                dense_hidden_dims=[4, 5],
                n_classes=4,
            ),
            466,
        ),
    ],
)
def test_count_params(model, expected_size):
    counts = count_params(model)
    assert counts["total_params"] == expected_size
    assert counts["trainable_params"] == expected_size
    assert counts["non_trainable_params"] == 0


def test_train_fn():
    torch.manual_seed(0)
    inputs = torch.randn(10, 3, 224, 224)
    outputs = torch.randint(0, 3, (10,))
    lr = 1e-4
    n_epochs = 10

    model = CNNClassifier((224, 224), 5, 3, [13, 14], [15, 16], n_classes=3).to(device)
    loss_fn = torch.nn.CrossEntropyLoss()
    iter_losses, epoch_losses = train_fn(
        model, loss_fn, input=inputs.to(device), output=outputs.to(device), lr=lr, epochs=n_epochs
    )

    assert epoch_losses[-1] < epoch_losses[0], "Loss should decrease"
    assert len(iter_losses) == n_epochs
    assert len(epoch_losses) == n_epochs

    # without verbose
    iter_losses, epoch_losses = train_fn(
        model, loss_fn, input=inputs.to(device), output=outputs.to(device), lr=lr, epochs=n_epochs, verbose=False
    )

    assert epoch_losses[-1] < epoch_losses[0], "Loss should decrease"
    assert len(iter_losses) == n_epochs
    assert len(epoch_losses) == n_epochs

    # wihout shuffle
    iter_losses, epoch_losses = train_fn(
        model, loss_fn, input=inputs.to(device), output=outputs.to(device), lr=lr, epochs=n_epochs, shuffle=False
    )

    assert epoch_losses[-1] < epoch_losses[0], "Loss should decrease"
    assert len(iter_losses) == n_epochs
    assert len(epoch_losses) == n_epochs


def test_train_fn_kl_minimize():
    class TrainableNormal(nn.Module):
        def __init__(self):
            super().__init__()
            self.loc = nn.Parameter(torch.tensor(0.0))
            self.scale = nn.Parameter(torch.tensor(1.0))

        def forward(self, input):
            return dist.Normal(self.loc, self.scale)

    def loss_fn(model_output, output, target_dist):
        learned_distribution = model_output
        return dist.kl_divergence(learned_distribution, target_dist)

    model = TrainableNormal()
    target_dist = dist.Normal(2.0, 3.0)

    # Tune without optimizer
    iter_losses, epoch_losses = train_fn(
        model, loss_fn, input=None, output=None, lr=0.1, epochs=50, loss_fn_kwargs={"target_dist": target_dist}
    )
    assert (model.loc - target_dist.loc).abs() < 0.2
    assert (model.scale - target_dist.scale).abs() < 0.2

    # Tune with optimizer
    optim = torch.optim.Adam(model.parameters(), lr=0.1)
    iter_losses, epoch_losses = train_fn(
        model, loss_fn, input=None, output=None, epochs=50, loss_fn_kwargs={"target_dist": target_dist}, optimizer=optim
    )

    # Return state dict
    (iter_losses, epoch_losses), state_dict_list = train_fn(
        model,
        loss_fn,
        input=None,
        output=None,
        lr=0.01,
        epochs=50,
        loss_fn_kwargs={"target_dist": target_dist},
        return_state_dict=True,
    )
    locs = torch.tensor([state_dict["loc"] for state_dict in state_dict_list])
    scales = torch.tensor([state_dict["scale"] for state_dict in state_dict_list])


def test_train_fn_wihout_outputs():
    torch.manual_seed(0)
    input = torch.randn(10, 3)

    model = MLPRegressor(input_dim=3, hidden_dims=[4, 5], output_dim=1).to(device)

    def loss_fn(model_output, output):
        return model_output.mean()

    iter_losses, epoch_losses = train_fn(model, loss_fn, input.to(device), output=None, lr=1e-4, epochs=10)

    assert epoch_losses[-1] < epoch_losses[0], "Loss should decrease"


def test_train_fn_with_dataloader():
    torch.manual_seed(0)
    input = torch.randn(10, 3)
    output = torch.randn(10, 1)

    model = MLPRegressor(input_dim=3, hidden_dims=[4, 5], output_dim=1).to(device)

    def loss_fn(model_output, output):
        return model_output.mean()

    from torch.utils.data import TensorDataset, DataLoader

    dataset = TensorDataset(input, output)
    dataloader = DataLoader(dataset, batch_size=2, shuffle=True)

    iter_losses, epoch_losses = train_fn(model, loss_fn, dataloader=dataloader, lr=1e-4, epochs=10)

    assert epoch_losses[-1] < epoch_losses[0], "Loss should decrease"



from abc import ABC, abstractmethod
import numpy as np
import torch
import torch.nn as nn
from torch.utils.data import DataLoader, TensorDataset
from typing import Sequence, Dict


# Base classes for acquisition functions


class Acquisition(ABC):
    @abstractmethod
    def acquire_scores(self, *args, **kwargs):
        raise NotImplementedError("This method must be implemented in a subclass")


class DeterministicAcquisition(Acquisition):
    @abstractmethod
    def acquire_scores(self, *args, **kwargs):
        raise NotImplementedError("This method must be implemented in a subclass")


class RandomAcquisition(Acquisition):
    @abstractmethod
    def acquire_scores(self, *args, **kwargs):
        raise NotImplementedError("This method must be implemented in a subclass")


class MCAcquisition(Acquisition):
    @abstractmethod
    def acquire_scores(self, *args, **kwargs):
        raise NotImplementedError("This method must be implemented in a subclass")


class EnsembleAcquisition(Acquisition):
    @abstractmethod
    def acquire_scores(self, *args, **kwargs):
        raise NotImplementedError("This method must be implemented in a subclass")


class DiversityAcquisition(Acquisition):
    @abstractmethod
    def acquire_scores(self, *args, **kwargs):
        raise NotImplementedError("This method must be implemented in a subclass")



import numpy as np

import torch
import torch.nn as nn
from torch.utils.data import DataLoader

from astra.torch.al.strategies.base import Strategy
from astra.torch.al.acquisitions.base import EnsembleAcquisition
from astra.torch.al.errors import AcquisitionMismatchError

from typing import Sequence, Dict, Union


class EnsembleStrategy(Strategy):
    def __init__(
        self,
        acquisitions: Union[EnsembleAcquisition, Sequence[EnsembleAcquisition]],
        inputs: torch.Tensor,
        outputs: torch.Tensor,
    ):
        """Ensemble query strategy with multiple neural networks.

        Args:
            acquisitions: A sequence of acquisition functions.
            inputs: A tensor of inputs.
            outputs: A tensor of outputs.
        """
        super().__init__(acquisitions, inputs, outputs)

        for name, acquisition in self.acquisitions.items():
            if not isinstance(acquisition, EnsembleAcquisition):
                raise AcquisitionMismatchError(EnsembleAcquisition, acquisition)

    def query(
        self,
        net: Sequence[torch.Tensor],
        pool_indices: torch.Tensor,
        context_indices: Sequence[int] = None,
        n_query_samples: int = 1,
        n_mc_samples: int = None,
        batch_size: int = None,
    ) -> Dict[str, torch.Tensor]:
        """
        Args:
            net: A sequence of neural networks.
            pool_indices: The indices of the pool set.
            context_indices: This argument is ignored.
            n_query_samples: Number of samples to query.
            n_mc_samples: This argument is ignored.
            batch_size: Batch size for the data loader.

        Returns:
            best_indices: A dictionary of acquisition names and the corresponding best indices.
        """
        assert isinstance(pool_indices, torch.Tensor), f"pool_indices must be a torch.Tensor, got {type(pool_indices)}"

        if not isinstance(net, Sequence):
            raise ValueError(f"net must be a sequence of nets, got {type(net)}")

        if batch_size is None:
            batch_size = len(pool_indices)

        data_loader = DataLoader(self.dataset[pool_indices])

        # Put model on eval mode
        for model in net:
            model.eval()

        with torch.no_grad():
            logits_list = []
            for x, _ in data_loader:
                net_logits_list = []
                for model in net:
                    net_logits = model(x.to(self.device))[np.newaxis, ...]
                    net_logits_list.append(net_logits)
                logits = torch.cat(net_logits_list, dim=0)  # (n_nets, batch_dim, n_classes)
                logits_list.append(logits)
            logits = torch.cat(logits_list, dim=1)  # (n_nets, pool_dim, n_classes)

            best_indices = {}
            for acq_name, acquisition in self.acquisitions.items():
                scores = acquisition.acquire_scores(logits)
                index = torch.topk(scores, n_query_samples).indices
                selected_indices = pool_indices[index]
                best_indices[acq_name] = selected_indices

        return best_indices



import wandb
from tqdm import tqdm
import torch
import optree
from torch.utils.data import TensorDataset, DataLoader
from typing import Optional


def count_params(model):
    params = [(p.numel(), True) if p.requires_grad else (p.numel(), False) for p in model.parameters()]
    sum_fn = lambda params: sum(map(lambda x: x[0], params))
    total_params = sum_fn(params)
    trainable_params = sum_fn(filter(lambda x: x[1] is True, params))
    non_trainable_params = sum_fn(filter(lambda x: x[1] is False, params))
    return {
        "total_params": total_params,
        "trainable_params": trainable_params,
        "non_trainable_params": non_trainable_params,
    }


def get_model_device(model):
    return next(model.parameters()).device


def train_fn(
    model: torch.nn.Module,
    loss_fn: callable,
    input: Optional[torch.Tensor] = None,
    output: Optional[torch.Tensor] = None,
    dataloader: Optional[DataLoader] = None,
    lr: Optional[float] = 0.01,
    epochs: int = 1,
    batch_size: Optional[int] = None,
    optimizer: Optional[torch.optim.Optimizer] = None,
    model_kwargs: dict = {},
    loss_fn_kwargs: dict = {},
    shuffle: bool = False,
    verbose: bool = True,
    return_state_dict: bool = False,
    wandb_log: bool = False,
):
    """Train a model with a loss function.

    Args:
        model (torch.nn.Module): A PyTorch model. It is called with `input` and `**model_kwargs`.
        loss_fn (callable): A callable loss function. First argument is the model output. Second argument is the `output` argument. Third argument is `**loss_fn_kwargs`.
        input (Optional[torch.Tensor]): It is passed as a first argument to the `model`.
        output (Optional[torch.Tensor]): It is passed as a second argument to the `loss_fn`. First argument to `loss_fn` is the model output.
        dataloader (Optional[DataLoader]): DataLoader for iterating through the data. If provided, it overrides `input` and `output`.
        lr (Optional[float]): Learning rate. It is ignored if `optimizer` is explicitly passed.
        epochs (int): Number of epochs.
        optimizer (Optional[torch.optim.Optimizer]): A PyTorch optimizer. If None, Adam optimizer with user specified `lr` is used.
        batch_size (Optional[int]): Batch size. If None, it is set to the length of `input` or `output` (whichever is not None). If both are None, it is set to 1.
        model_kwargs (dict): Keyword arguments passed to the `model`.
        loss_fn_kwargs (dict): Keyword arguments passed to the `loss_fn`.
        shuffle (bool): If True, shuffle the data **before** each epoch.
        verbose (bool): If True, print loss values and progress with `tqdm`.
        return_state_dict (bool): If True, store state_dicts of the model at the end of each iteration and return them as a list in addition to the losses.
        wandb_log (bool): If True, log the loss values to wandb.

    Returns:
        If `return_state_dict` is False:
            (iter_losses, epoch_losses)
        Else:
            ((iter_losses, epoch_losses), state_dict_history)
    """

    device = get_model_device(model)

    def get_batch():
        if input is None and output is None:
            yield None, None
            return 0  # execution successfully finished

        in_or_out = output if input is None else input
        if batch_size is None:
            inner_batch_size = len(in_or_out)
        else:
            inner_batch_size = batch_size

        iterable = range(0, len(in_or_out), inner_batch_size)

        if shuffle:
            idx = torch.randperm(len(in_or_out))
        else:
            idx = torch.arange(len(in_or_out))

        for i in iterable:
            if input is not None:
                batch_input = input[idx[i : i + inner_batch_size]]
            else:
                batch_input = None

            if output is not None:
                batch_output = output[idx[i : i + inner_batch_size]]
            else:
                batch_output = None

            yield batch_input, batch_output

    def one_step(batch_input, batch_output):
        optimizer.zero_grad()
        model_output = model(batch_input, **model_kwargs)
        loss = loss_fn(model_output, batch_output, **loss_fn_kwargs)
        loss.backward()
        optimizer.step()
        loss_item = loss.item()
        return loss_item

    model.train()
    if optimizer is None:
        optimizer = torch.optim.Adam(model.parameters(), lr=lr)

    iter_losses = []
    epoch_losses = []
    state_dict_history = []

    pbar = range(epochs)
    if verbose:
        pbar = tqdm(pbar)

    if dataloader is not None:
        inner_loop = lambda: dataloader
    else:
        inner_loop = lambda: get_batch()

    for _ in pbar:
        loss_buffer = []
        for batch_input, batch_output in inner_loop():
            batch_input = batch_input.to(device) if isinstance(batch_input, torch.Tensor) else batch_input
            batch_output = batch_output.to(device) if isinstance(batch_output, torch.Tensor) else batch_output

            loss = one_step(batch_input, batch_output)
            if wandb_log:
                wandb.log({"loss": loss})
            iter_losses.append(loss)
            loss_buffer.append(loss)

            if return_state_dict:
                # state_dict history should be on CPU to avoid unnecessary clogging of GPU memory
                state_dict_history.append(optree.tree_map(lambda x: x.cpu(), model.state_dict()))

        epoch_loss = sum(loss_buffer) / len(loss_buffer)
        epoch_losses.append(epoch_loss)

        if verbose:
            pbar.set_description(f"Loss: {epoch_loss:.8f}")

    if return_state_dict:
        return (iter_losses, epoch_losses), state_dict_history
    else:
        return iter_losses, epoch_losses



from copy import deepcopy
import os
import numpy as np
import torch
from torchvision import datasets, transforms
import warnings


def prerequisite(f):
    if "TORCH_HOME" not in os.environ:
        os.environ["TORCH_HOME"] = os.path.expanduser("~/.cache/torch")
        warnings.warn(f"TORCH_HOME not set, setting it to {os.environ['TORCH_HOME']}")
    return f


@prerequisite
def load_mnist():
    mnist_train = datasets.MNIST(root=f"{os.environ['TORCH_HOME']}/data", train=True, download=True)
    mnist_test = datasets.MNIST(root=f"{os.environ['TORCH_HOME']}/data", train=False, download=True)

    train_images = mnist_train.data.float() / 255
    train_labels = mnist_train.targets.long()

    test_images = mnist_test.data.float() / 255
    test_labels = mnist_test.targets.long()

    images = torch.cat([train_images, test_images], dim=0)
    labels = torch.cat([train_labels, test_labels], dim=0)

    mnist_train.data = images
    mnist_train.targets = labels

    def repr(self):
        return f"""
MNIST Dataset
length of dataset: {len(self)}
shape of images: {self.data.shape[1:]}
len of classes: {len(self.classes)}
classes: {self.classes}
dtype of images: {self.data.dtype}
dtype of labels: {self.targets.dtype}
range of image values: min={self.data.min()}, max={self.data.max()}
"""

    mnist_train.__class__.__repr__ = repr

    return mnist_train


@prerequisite
def load_cifar_10():
    cfar_10_train = datasets.CIFAR10(root=f"{os.environ['TORCH_HOME']}/data", train=True, download=True)
    cfar_10_test = datasets.CIFAR10(root=f"{os.environ['TORCH_HOME']}/data", train=False, download=True)

    train_images = torch.tensor(cfar_10_train.data).float() / 255
    train_images = torch.einsum("nhwc->nchw", train_images)
    train_labels = torch.tensor(cfar_10_train.targets).long()

    test_images = torch.tensor(cfar_10_test.data).float() / 255
    test_images = torch.einsum("nhwc->nchw", test_images)
    test_labels = torch.tensor(cfar_10_test.targets).long()

    images = torch.cat([train_images, test_images], dim=0)
    labels = torch.cat([train_labels, test_labels], dim=0)

    cfar_10_train.data = images
    cfar_10_train.targets = labels

    def repr(self):
        return f"""
CIFAR-10 Dataset
length of dataset: {len(self)}
shape of images: {self.data.shape[1:]}
len of classes: {len(self.classes)}
classes: {self.classes}
dtype of images: {self.data.dtype}
dtype of labels: {self.targets.dtype}
range of image values: min={self.data.min()}, max={self.data.max()}
            """

    cfar_10_train.__class__.__repr__ = repr

    return cfar_10_train



from abc import ABC, abstractmethod
import numpy as np
import torch
import torch.nn as nn
from torch.utils.data import DataLoader, TensorDataset
from typing import Sequence, Dict

from astra.torch.al.acquisitions.base import Acquisition

# Base class for query strategies


class Strategy(ABC, nn.Module):
    def __init__(
        self,
        acquisitions: Sequence[Acquisition],
        inputs: torch.Tensor,
        outputs: torch.Tensor,
    ):
        """Base class for query strategies

        Args:
            acquisitions: A sequence of acquisition functions.
            inputs: A tensor of inputs.
            outputs: A tensor of outputs.
        """
        super().__init__()
        assert isinstance(inputs, torch.Tensor), f"inputs must be a torch.Tensor, got {type(inputs)}"
        assert isinstance(outputs, torch.Tensor), f"outputs must be a torch.Tensor, got {type(outputs)}"

        self.inputs = inputs
        self.outputs = outputs
        self.n_classes = len(torch.unique(outputs))
        self.dataset = TensorDataset(self.inputs, self.outputs)
        self.dummy_param = nn.Parameter(torch.zeros(1))

        if not isinstance(acquisitions, Sequence):
            acquisitions = [acquisitions]
        self.acquisitions = {acq.__class__.__name__: acq for acq in acquisitions}

    @property
    def device(self):
        return self.dummy_param.device

    @abstractmethod
    def query(self, *args, **kwargs):
        raise NotImplementedError("This method must be implemented in a subclass")



import torch
import torch.nn as nn
from torch.utils.data import DataLoader

from astra.torch.al.strategies.base import Strategy
from astra.torch.al.acquisitions.base import DiversityAcquisition
from astra.torch.al.errors import AcquisitionMismatchError

from typing import Sequence, Dict, Union


class DiversityStrategy(Strategy):
    def __init__(
        self,
        acquisitions: Union[DiversityAcquisition, Sequence[DiversityAcquisition]],
        inputs: torch.Tensor,
        outputs: torch.Tensor,
    ):
        """Diversity query strategy to maximize the diversity of the selected samples.

        Args:
            acquisitions: A sequence of acquisition functions.
            inputs: A tensor of inputs.
            outputs: A tensor of outputs.
        """
        super().__init__(acquisitions, inputs, outputs)

        for name, acquisition in self.acquisitions.items():
            if not isinstance(acquisition, DiversityAcquisition):
                raise AcquisitionMismatchError(DiversityAcquisition, acquisition)

    def query(
        self,
        net: nn.Module,
        pool_indices: torch.Tensor,
        context_indices: torch.Tensor = None,
        n_query_samples: int = 1,
        n_mc_samples: int = None,
        batch_size: int = None,
    ) -> Dict[str, torch.Tensor]:
        """
        Args:
            net: A neural network to extract features.
            pool_indices: The indices of the pool set.
            context_indices: The indices of the context set which is used to compute the diversity.
            n_query_samples: Number of samples to query.
            n_mc_samples: This argument is ignored.
            batch_size: Batch size for the data loader.

        Returns:
            best_indices: A dictionary of acquisition names and the corresponding best indices.
        """
        assert isinstance(pool_indices, torch.Tensor), f"pool_indices must be a torch.Tensor, got {type(pool_indices)}"
        assert isinstance(
            context_indices, torch.Tensor
        ), f"context_indices must be a torch.Tensor, got {type(context_indices)}"

        if batch_size is None:
            batch_size = len(pool_indices)

        data_loader = DataLoader(self.dataset)

        # put model on eval mode
        net.eval()

        with torch.no_grad():
            # Get all features
            features_list = []
            for x, _ in data_loader:
                features = net(x)
                features_list.append(features)
            features = torch.cat(features_list, dim=0)  # (data_dim, feature_dim)

            best_indices = {}

            if not isinstance(pool_indices, list):
                # tolist() works for both numpy and torch tensors. It also work for tensors on GPU.
                pool_indices = pool_indices.tolist()
            if not isinstance(context_indices, list):
                context_indices = context_indices.tolist()

            for acq_name, acquisition in self.acquisitions.items():
                selected_indices = []
                # TODO: We can make this loop faster by computing scores only for updated indices. There can be a method in acquisition to update the scores.
                for _ in range(n_query_samples):
                    scores = acquisition.acquire_scores(features, pool_indices, context_indices)
                    index = torch.argmax(scores)
                    selected_index = pool_indices[index]
                    selected_indices.append(selected_index)
                    pool_indices = torch.cat([pool_indices[:index], pool_indices[index + 1 :]])
                    context_indices = torch.cat([context_indices, selected_index])
                selected_indices = torch.tensor(selected_indices, device=self.device)
                best_indices[acq_name] = selected_indices

        return best_indices



import numpy as np

import torch
import torch.nn as nn
from torch.utils.data import DataLoader

from astra.torch.al.strategies.base import Strategy
from astra.torch.al.acquisitions.base import MCAcquisition
from astra.torch.al.errors import AcquisitionMismatchError

from typing import Sequence, Dict, Union


class MCStrategy(Strategy):
    def __init__(
        self,
        acquisitions: Union[MCAcquisition, Sequence[MCAcquisition]],
        inputs: torch.Tensor,
        outputs: torch.Tensor,
    ):
        """Monte Carlo Dropout query strategy. See https://arxiv.org/abs/1506.02142 (Dropout as a Bayesian Approximation: Representing Model Uncertainty in Deep Learning)

        Args:
            acquisitions: A sequence of acquisition functions.
            inputs: A tensor of inputs.
            outputs: A tensor of outputs.
        """
        super().__init__(acquisitions, inputs, outputs)

        for name, acquisition in self.acquisitions.items():
            if not isinstance(acquisition, MCAcquisition):
                raise AcquisitionMismatchError(MCAcquisition, acquisition)

    def query(
        self,
        net: nn.Module,
        pool_indices: torch.Tensor,
        context_indices: torch.Tensor = None,
        n_query_samples: int = 1,
        n_mc_samples: int = 10,
        batch_size: int = None,
    ) -> Dict[str, torch.Tensor]:
        """
        Args:
            net: A neural network with dropout layers.
            pool_indices: The indices of the pool set.
            context_indices: This argument is ignored.
            n_query_samples: Number of samples to query.
            n_mc_samples: Number of Monte Carlo samples to draw from the model.
            batch_size: Batch size for the data loader.

        Returns:
            best_indices: A dictionary of acquisition names and the corresponding best indices.
        """
        assert isinstance(pool_indices, torch.Tensor), f"pool_indices must be a torch.Tensor, got {type(pool_indices)}"

        if batch_size is None:
            batch_size = len(pool_indices)

        # Set n_repeats for x for vmap later
        repeats = [1] * (len(self.dataset.tensor[0].shape) + 1)  # +1 for mc dimension
        repeats[0] = n_mc_samples

        data_loader = DataLoader(self.dataset[pool_indices])

        # Put the model on train mode to enable dropout
        net.train()

        with torch.no_grad():
            logits_list = []
            for x, _ in data_loader:
                vx = x[np.newaxis, ...].repeat(*repeats).to(self.device)
                logits = torch.vmap(net, randomness="different")(vx)
                logits_list.append(logits)
            logits = torch.cat(logits_list, dim=1)  # (n_mc_samples, pool_dim, n_classes)

            best_indices = {}
            for acq_name, acquisition in self.acquisitions.items():
                scores = acquisition.acquire_scores(logits)
                index = torch.topk(scores, n_query_samples).indices
                selected_indices = pool_indices[index]
                best_indices[acq_name] = selected_indices

        return best_indices



from setuptools import find_packages, setup

with open("requirements.txt") as f:
    requirements = f.read().strip().split("\n")

setup(
    packages=find_packages(),
    python_requires=">=3.9",
    install_requires=requirements,
    include_package_data=True,
)



import pytest

import torch
from torchvision.datasets import CIFAR10

from astra.torch.models import CNNClassifier
from astra.torch.al import UniformRandomAcquisition, RandomStrategy, EnsembleAcquisition

device = torch.device("cuda" if torch.cuda.is_available() else "cpu")


def test_random():
    data = CIFAR10(root="data", download=True, train=False)  # "test" for less data
    inputs = torch.tensor(data.data).float().to(device)
    outputs = torch.tensor(data.targets).long().to(device)

    # Meta parameters
    n_pool = 1000
    indices = torch.randperm(len(inputs))
    pool_indices = indices[:n_pool]
    train_indices = indices[n_pool:]
    n_query_samples = 10

    # Define the acquisition function
    acquisition = UniformRandomAcquisition()
    strategy = RandomStrategy(acquisition, inputs, outputs)

    # Put the strategy on the device
    strategy.to(device)

    # Define the model
    net = CNNClassifier((32, 32), 3, 3, [4, 8], [2, 3], 10).to(device)

    # Query the strategy
    best_indices = strategy.query(net, pool_indices, n_query_samples=n_query_samples)

    assert best_indices["UniformRandomAcquisition"].shape == (n_query_samples,)



import pytest
import torch
import astra.torch.models as models
from torchvision.models import (
    efficientnet_b0,
    EfficientNet_B0_Weights,
    vit_b_16,
    ViT_B_16_Weights,
    resnet18,
    ResNet18_Weights,
    alexnet,
    AlexNet_Weights,
    densenet121,
    DenseNet121_Weights,
)

device = "cuda" if torch.cuda.is_available() else "cpu"


@pytest.mark.parametrize(
    "x, output_dim, out_shape",
    [
        (torch.randn(11, 3), 4, torch.Size([11, 4])),
        (torch.randn(12, 5), 6, torch.Size([12, 6])),
    ],
)
def test_mlp(x, output_dim, out_shape):
    # create a empty class object

    input_dim = x.shape[1]
    model = models.MLPRegressor(input_dim, [7, 9], output_dim).to(device)

    output = model(x.to(device))
    assert output.shape == out_shape

    model = models.MLPClassifier(input_dim, [7, 9], output_dim).to(device)
    output = model(x.to(device))
    assert output.shape == torch.Size([x.shape[0], output_dim])

    pred_logits = model.predict(x.to(device), batch_size=2)
    pred_classes = model.predict_class(x.to(device), batch_size=2)
    assert pred_logits.shape == torch.Size([x.shape[0], output_dim])
    assert pred_classes.shape == torch.Size([x.shape[0]])
    assert torch.all(pred_logits.argmax(dim=1) == pred_classes)
    assert pred_logits.dtype == torch.float32
    assert pred_classes.dtype == torch.int64


@pytest.mark.parametrize(
    "x, output_dim, out_shape",
    [
        (torch.randn(11, 3), 4, torch.Size([11, 4])),
        (torch.randn(12, 5), 6, torch.Size([12, 6])),
    ],
)
def test_siren(x, output_dim, out_shape):
    # create a empty class object

    input_dim = x.shape[1]
    model = models.SIRENRegressor(input_dim, [7, 9], output_dim).to(device)

    output = model(x.to(device))
    assert output.shape == out_shape


@pytest.mark.parametrize("x", [torch.randn(11, 3, 44, 28), torch.randn(11, 5, 224, 180)])
def test_cnn(x):
    batch_dim = x.shape[0]
    image_dim = (x.shape[-2], x.shape[-1])
    n_channels = x.shape[1]
    kernel_size = 3
    conv_hidden_dims = [4, 4, 7]
    dense_hidden_dims = [7, 9]
    output_dim = 12
    dropout = 0.1

    model = models.CNNClassifier(
        image_dim, kernel_size, n_channels, conv_hidden_dims, dense_hidden_dims, output_dim, dropout=dropout
    ).to(device)
    out = model(x.to(device))

    assert out.shape == torch.Size([batch_dim, output_dim])

    # With aggregation
    model = models.CNNClassifier(
        image_dim, kernel_size, n_channels, conv_hidden_dims, dense_hidden_dims, output_dim, adaptive_pooling=True
    ).to(device)
    out = model(x.to(device))

    assert out.shape == torch.Size([batch_dim, output_dim])

    model = models.CNNRegressor(
        image_dim, kernel_size, n_channels, conv_hidden_dims, dense_hidden_dims, output_dim, dropout=dropout
    ).to(device)
    out = model(x.to(device))

    assert out.shape == torch.Size([batch_dim, output_dim])

    # With aggregation
    model = models.CNNRegressor(
        image_dim, kernel_size, n_channels, conv_hidden_dims, dense_hidden_dims, output_dim, adaptive_pooling=True
    ).to(device)
    out = model(x.to(device))

    assert out.shape == torch.Size([batch_dim, output_dim])


@pytest.mark.parametrize("x, output_dim", [(torch.randn(11, 3, 28, 28), 4), (torch.randn(11, 3, 224, 224), 6)])
def test_resnet(x, output_dim):
    batch_dim = x.shape[0]

    model = models.ResNetClassifier(resnet18, ResNet18_Weights.DEFAULT, output_dim).to(device)
    out = model(x.to(device))

    assert out.shape == torch.Size([batch_dim, output_dim])

    model = models.ResNetRegressor(resnet18, ResNet18_Weights.DEFAULT, output_dim).to(device)
    out = model(x.to(device))

    assert out.shape == torch.Size([batch_dim, output_dim])


@pytest.mark.parametrize("x, output_dim", [(torch.randn(11, 3, 28, 28), 4), (torch.randn(11, 3, 224, 224), 6)])
def test_efficientnet(x, output_dim):
    batch_dim = x.shape[0]

    model = models.EfficientNetClassifier(efficientnet_b0, EfficientNet_B0_Weights.DEFAULT, output_dim).to(device)
    out = model(x.to(device))

    assert out.shape == torch.Size([batch_dim, output_dim])

    model = models.EfficientNetRegressor(efficientnet_b0, EfficientNet_B0_Weights.DEFAULT, output_dim).to(device)
    out = model(x.to(device))

    assert out.shape == torch.Size([batch_dim, output_dim])


@pytest.mark.parametrize("x, output_dim", [(torch.randn(11, 3, 28, 28), 4), (torch.randn(11, 3, 224, 224), 6)])
def test_vit(x, output_dim):
    batch_dim = x.shape[0]

    if x.shape[-1] != 224 or x.shape[-2] != 224:
        model = models.ViTClassifier(vit_b_16, ViT_B_16_Weights.DEFAULT, output_dim, transform=False).to(device)

        with pytest.raises(AssertionError):
            out = model(x.to(device))

    model = models.ViTClassifier(vit_b_16, ViT_B_16_Weights.DEFAULT, output_dim, transform=True).to(device)

    out = model(x.to(device))

    assert out.shape == torch.Size([batch_dim, output_dim])

    if x.shape[-1] != 224 or x.shape[-2] != 224:
        model = models.ViTRegressor(vit_b_16, ViT_B_16_Weights.DEFAULT, output_dim, transform=False).to(device)

        with pytest.raises(AssertionError):
            out = model(x.to(device))

    model = models.ViTRegressor(vit_b_16, ViT_B_16_Weights.DEFAULT, output_dim, transform=True).to(device)

    out = model(x.to(device))

    assert out.shape == torch.Size([batch_dim, output_dim])


@pytest.mark.parametrize("x, output_dim", [(torch.randn(11, 3, 512, 246), 4), (torch.randn(11, 3, 224, 224), 6)])
def test_alexnet(x, output_dim):
    batch_dim = x.shape[0]

    model = models.AlexNetClassifier(alexnet, AlexNet_Weights.DEFAULT, output_dim).to(device)
    out = model(x.to(device))

    assert out.shape == torch.Size([batch_dim, output_dim])

    model = models.AlexNetRegressor(alexnet, AlexNet_Weights.DEFAULT, output_dim).to(device)
    out = model(x.to(device))

    assert out.shape == torch.Size([batch_dim, output_dim])


@pytest.mark.parametrize("x, output_dim", [(torch.randn(11, 3, 512, 246), 4), (torch.randn(11, 3, 224, 224), 6)])
def test_densenet(x, output_dim):
    batch_dim = x.shape[0]

    model = models.DenseNetClassifier(densenet121, DenseNet121_Weights.DEFAULT, output_dim).to(device)
    out = model(x.to(device))

    assert out.shape == torch.Size([batch_dim, output_dim])

    model = models.DenseNetRegressor(densenet121, DenseNet121_Weights.DEFAULT, output_dim).to(device)
    out = model(x.to(device))

    assert out.shape == torch.Size([batch_dim, output_dim])



from astra.torch.utils import count_params
from astra.torch.models import MLPRegressor

mlp = MLPRegressor(input_dim=2, hidden_dims=[5, 6], output_dim=1)

n_params = count_params(mlp)
print(n_params)
